@@ -19,6 +19,29 @@ __all__ = ('Nvim')
 
 os_chdir = os.chdir
 
+lua_module = """
+local a = vim.api
+local function update_highlights(buf, src_id, hls, clear_first, clear_end)
+  if clear_first ~= nil then
+      a.nvim_buf_clear_highlight(buf, src_id, clear_first, clear_end)
+  end
+  for _,hl in pairs(hls) do
+    local group, line, col_start, col_end = unpack(hl)
+    if col_start == nil then
+      col_start = 0
+    end
+    if col_end == nil then
+      col_end = -1
+    end
+    a.nvim_buf_add_highlight(buf, src_id, group, line, col_start, col_end)
+  end
+end
+
+local chid = ...
+local mod = {update_highlights=update_highlights}
+_G["_pynvim_"..chid] = mod
+"""
+
 
 class Nvim(object):
 
@@ -92,6 +115,7 @@ class Nvim(object):
         self.current = Current(self)
         self.session = CompatibilitySession(self)
         self.funcs = Funcs(self)
+        self.lua = LuaFuncs(self)
         self.error = NvimError
         self._decode = decode
         self._err_cb = err_cb
@@ -115,6 +139,12 @@ class Nvim(object):
             return ExtType(*obj.code_data)
         return obj
 
+    def _get_lua_private(self):
+        if not getattr(self._session, "_has_lua", False):
+            self.exec_lua(lua_module, self.channel_id)
+            self._session._has_lua = True
+        return getattr(self.lua, "_pynvim_{}".format(self.channel_id))
+
     def request(self, name, *args, **kwargs):
         r"""Send an API request or notification to nvim.
 
@@ -122,16 +152,16 @@ class Nvim(object):
         functions have python wrapper functions. The `api` object can
         be also be used to call API functions as methods:
 
-            vim.api.err_write('ERROR\n', async=True)
+            vim.api.err_write('ERROR\n', async_=True)
             vim.current.buffer.api.get_mark('.')
 
         is equivalent to
 
-            vim.request('nvim_err_write', 'ERROR\n', async=True)
+            vim.request('nvim_err_write', 'ERROR\n', async_=True)
             vim.request('nvim_buf_get_mark', vim.current.buffer, '.')
 
 
-        Normally a blocking request will be sent.  If the `async` flag is
+        Normally a blocking request will be sent.  If the `async_` flag is
         present and True, a asynchronous notification is sent instead. This
         will never block, and the return value or error is ignored.
         """
@@ -253,6 +283,27 @@ class Nvim(object):
         """Call a vimscript function."""
         return self.request('nvim_call_function', name, args, **kwargs)
 
+    def exec_lua(self, code, *args, **kwargs):
+        """Execute lua code.
+
+        Additional parameters are available as `...` inside the lua chunk.
+        Only statements are executed.  To evaluate an expression, prefix it
+        with `return`: `return my_function(...)`
+
+        There is a shorthand syntax to call lua functions with arguments:
+
+            nvim.lua.func(1,2)
+            nvim.lua.mymod.myfunction(data, async_=True)
+
+        is equivalent to
+
+            nvim.exec_lua("return func(...)", 1, 2)
+            nvim.exec_lua("mymod.myfunction(...)", data, async_=True)
+
+        Note that with `async_=True` there is no return value.
+        """
+        return self.request('nvim_execute_lua', code, args, **kwargs)
+
     def strwidth(self, string):
         """Return the number of display cells `string` occupies.
 
@@ -467,5 +518,29 @@ class Funcs(object):
         return partial(self._nvim.call, name)
 
 
+class LuaFuncs(object):
+
+    """Wrapper to allow lua functions to be called like python methods."""
+
+    def __init__(self, nvim, name=""):
+        self._nvim = nvim
+        self.name = name
+
+    def __getattr__(self, name):
+        """Return wrapper to named api method."""
+        prefix = self.name + "." if self.name else ""
+        return LuaFuncs(self._nvim, prefix + name)
+
+    def __call__(self, *args, **kwargs):
+        # first new function after keyword rename, be a bit noisy
+        if 'async' in kwargs:
+            raise ValueError('"async" argument is not allowed. '
+                             'Use "async_" instead.')
+        async_ = kwargs.get('async_', False)
+        pattern = "return {}(...)" if not async_ else "{}(...)"
+        code = pattern.format(self.name)
+        return self._nvim.exec_lua(code, *args, **kwargs)
+
+
 class NvimError(Exception):
     pass
